#!/vol/PACK/perl5.004_04/bin/perl -w
#
# ConnWeb.pm
#
# Andrew.Findlay@brunel.ac.uk
# August 1999
#
# @(#) ConnWeb.pm 1.58@(#)

package ConnWeb;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
	HandleCGI
	SetHTMLDir
	SetDBParamFile
	SetCgiUrl
	SetAllowFudgeUser
	SetTestMode
	SetSaveQueries
	MailNotifications
);


use strict;
use ConnDB;
use ConnLog;
use HTML::Template;
use CGI;
use Mail::Send;
use Business::CreditCard;
# We use a local copy of ParseWords as the perl5.004_04 version is buggy
use ParseWords;

# $TempFile::TMPDIRECTORY = '/nosuch';


########################################################################

my ($MailFrom) = "networks (Networks Group)";

# Recipient of all mail sent in test mode
my ($TestRecipient) = "andrew";

########################################################################
# Where are the HTML template files kept?
########################################################################
my ($htmldir);

sub SetHTMLDir {
	($htmldir) = @_;
}

########################################################################
# Where is the database parameter file?
########################################################################
my ($database_param_file);

sub SetDBParamFile {
	($database_param_file) = @_;
}

########################################################################
# What is the URL to invoke the CGI script?
########################################################################
my ($cgi_url);

sub SetCgiUrl {
	($cgi_url) = @_;
}

########################################################################
# Should we bypass webserver authentication for testing?
########################################################################
my ($AllowFudgeUser) = 0;

sub SetAllowFudgeUser {
	($AllowFudgeUser) = @_;
}

########################################################################
# Are we in test mode?
########################################################################
my ($TestMode) = 0;

sub SetTestMode {
	($TestMode) = @_;
}

########################################################################
# Should we send mail?
########################################################################
my ($DoMail) = 0;

sub MailNotifications {
	($DoMail) = @_;
}

########################################################################
# Save queries to a logfile if this is set to a filename.
# Dont save if set to zero
########################################################################
# $SaveQueries = "/tmp/conn-queries";
my ($SaveQueries) = 0;

sub SetSaveQueries {
	($SaveQueries) = @_;
}

########################################################################
# Set parameters in the CGI module to guard against denial-of-service attack
########################################################################
$CGI::POST_MAX=1024 * 100;	# Limit CGI POST to 100k bytes
$CGI::DISABLE_UPLOADS = 0;	# Dont prevent CGI file uploads

########################################################################
# Globals for all operations - filled in by main dispatcher
########################################################################
my ($query, $mode, $op, $loggedinuser);

########################################################################
# Constants for state machine
########################################################################

my ($CONST_halls_service) = 1;
my ($CONST_departmental_service) = 2;

my ($CONST_srv_state_awaiting_payment) = 1;
my ($CONST_srv_state_awaiting_connection) = 2;
my ($CONST_srv_state_connected) = 3;
my ($CONST_srv_state_awaiting_disconnection) = 4;
my ($CONST_srv_state_disconnected) = 5;
my ($CONST_srv_state_cancelled) = 6;
my ($CONST_srv_state_awaiting_renewal_payment) = 7;


my ($CONST_semester_days_tolerance) = 14;

########################################################################
# Emit an HTML page with appropriate style
########################################################################
sub EmitPage {
	my ($title, $description, $body) = @_;

	my ($template);

	if ($TestMode) {
		$template = HTML::Template->new(
			filename => "$htmldir/testing-layout.html");
	}
	else {
		$template = HTML::Template->new(
			filename => "$htmldir/page-layout.html");
	}

	$template->param( "title", $title );
	$template->param( "description", $description );
	$template->param( "body", $body );

	print "Content-type: text/html\n\n";
	print $template->output();
}

########################################################################
# Emit an error page
########################################################################
sub EmitError {
	my ($message) = @_;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/error.html" );

	$template->param( "message", $message );
	EmitPage( "Error", "Error message", $template->output() );
}

########################################################################
# Emit a page of parameters (debugging utility)
########################################################################
sub ShowParams {
	# Security check
	if (!$TestMode) {
		Log( "err", "ShowParams called but not in test mode" );
		EmitError( "Permission denied" );
		exit 1;
	}

	my (@param) = $query->param;
	my ($er) = $query->cgi_error();
	if (!defined($er)) {
		$er = "UNDEFINED";
	}
	my ($infostring) = "cgi_error is $er<BR>";
	foreach (@param) {
		my ($val) = $query->param($_);
		$infostring .= "$_=$val<BR>";
	}
	$infostring .= "ENVIRONMENT:<BR>";
	foreach (keys(%ENV)) {
		$infostring .= "$_=$ENV{$_}<BR>";
	}
	#my ($qd) = $query->dump;
	#EmitError( "<B>Parameters received:</B><BR>$qd\n" );
	EmitError( "<B>Parameters received:</B><BR>$infostring" );
}

########################################################################
# Send a mail message: SendMail( <to>, <subject>, <message> )
########################################################################
sub SendMail {
        my ($to, $subject, $text) = @_;

	# Should we send mail at all?
	return if !$DoMail;

	$to = $TestRecipient if $TestMode;

        my ($msg) = new Mail::Send;
        $msg->to($to);
        $msg->subject($subject);
        $msg->set('From', $MailFrom);

        # We must use Sendmail transport to be able to set 'From'
        my ($fh) = $msg->open('sendmail');

        print $fh $text;
        $fh->close;
}


########################################################################
# Humanise a date string
# e.g. 2000-03-22 becomes "22 Mar 2000"
# Does no sanity checking of args
########################################################################

my (@MonthNames) = (
	"xxx",
	"Jan",
	"Feb",
	"Mar",
	"Apr",
	"May",
	"Jun",
	"Jul",
	"Aug",
	"Sep",
	"Oct",
	"Nov",
	"Dec"
);

sub HumaniseDate {
	my ($sysdate) = @_;

	# Pass empty strings straight through
	return "" if ($sysdate eq "");

	# Allow for timestamp format as well
	my ($datepart, $rest) = split /\s+/, $sysdate, 2;
	if (defined($rest)) {
		$rest = " $rest";
	}
	else {
		$rest = "";
	}

	my ($year, $month, $day) = split /-/, $sysdate;

	return sprintf "%d %s %d%s", $day, $MonthNames[$month], $year, $rest;
}


########################################################################
# Utility procedure to build a building-selector
########################################################################

sub BuildBuildingSelector {
	my ($building_spec) = @_;

	my ($buildings) = FetchBuildingList( $building_spec );
	if (!$buildings) {
		EmitError( "FetchBuildingList failed $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my (@item_hash_array);
	for ($item=0;defined($buildings->[$item]);$item++) {
		$item_hash_array[$item] = {
			"building_id" => $buildings->[$item][0],
			"building_name" => $buildings->[$item][2]
		}
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/building-selector-frag.html" );


	$template->param( "building", \@item_hash_array );

	return $template->output();
}

########################################################################
# Utility procedure to build a site-selector
########################################################################

sub BuildSiteSelector {
	my ($sites) = FetchSiteList;
	if (!$sites) {
		EmitError( "FetchSiteList failed $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my (@item_hash_array);
	for ($item=0;defined($sites->[$item]);$item++) {
		$item_hash_array[$item] = {
			"site_id" => $sites->[$item][0],
			"site_name" => $sites->[$item][1]
		}
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/site-selector-frag.html" );


	$template->param( "site", \@item_hash_array );

	return $template->output();
}


########################################################################
# Utility procedure to build a device-selector
# Includes only those devices that are available in the given
# building and block number.
# Name the selector using the request_id.
########################################################################

sub BuildDeviceSelector {
	my ($req_id, $building_id, $block_number) = @_;

	my ($devices) = FetchDeviceList( $building_id, $block_number );
	if (!$devices) {
		EmitError( "Internal error: FetchDeviceList failed $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my (@item_hash_array);
	for ($item=0;defined($devices->[$item]);$item++) {
		$item_hash_array[$item] = {
			"device_id" => $devices->[$item][0],
			"device_name" => $devices->[$item][1]
		}
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/device-selector-frag.html" );


	$template->param( "req_id", $req_id );
	$template->param( "devices", \@item_hash_array );

	return $template->output();
}

########################################################################
# Utility procedure to build a semester-selector
#
# If param 1 is the empty string, lists all current and future semesters.
# Otherwise, param 1 is a target start date and param 2 is a
# tolerance given in days.
########################################################################

sub BuildSemesterSelector {

	my ($target_start_date, $tolerance) = @_;

	my ($semesters);

	if ($target_start_date eq "") {
		$semesters = FetchSemesterList;
	}
	else {
		$semesters = FetchRenewalSemesterList($target_start_date, $tolerance);
	}
	if (!$semesters) {
		EmitError( "Internal error: FetchSemesterList failed $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my (@item_hash_array);

	for ($item=0;defined($semesters->[$item]);$item++) {
		# Connection fees are not paid again on renewal
		# We do not show start date on renewal selector
		my ($start_date, $conn_fee);
		if ($target_start_date eq "") {
			$conn_fee = $semesters->[$item][3];
			$start_date = $semesters->[$item][1];
		}
		else {
			$conn_fee = "0.00";
			$start_date = "";
		}

		$item_hash_array[$item] = {
			"sem_id" => $semesters->[$item][0],
			"sem_start_date" => HumaniseDate($start_date),
			"sem_end_date" => HumaniseDate($semesters->[$item][2]),
			"conn_fee" => $conn_fee,
			"sem_fee" => $semesters->[$item][4],
			"sem_descr" => $semesters->[$item][5]
		}
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/semester-selector-frag.html" );


	$template->param( "semester", \@item_hash_array );

	return $template->output();
}

########################################################################
# Emit an "apply for Halls Service" page
########################################################################
sub EmitApplyForHallsServicePage {

	my ($template) = HTML::Template->new(
		filename => "$htmldir/halls-application-form.html" );

	$template->param( "user", $loggedinuser );

	my ($fullname) = FetchFullName( $loggedinuser, 0 );
	if (!$fullname) {
		EmitError( "No entry found for user '$loggedinuser'" );
		Log( "err", "No entry found for user '$loggedinuser'" );
		return 1;
	}
	$template->param( "fullname", $fullname );

	$template->param( "cgi_url", $cgi_url );

	# We only want Halls on the building list
	my ($building_selector) = BuildBuildingSelector( 2 );
	$template->param( "building_selector", $building_selector );

	my ($semester_selector) = BuildSemesterSelector("", 0);
	$template->param( "semester_selector", $semester_selector );

	EmitPage( "Apply for network access in your Hall room",
		"Application form for network service in Halls of Residence",
		$template->output() );
}


########################################################################
# Emit a "proxy application for Halls Service" page
########################################################################
sub EmitProxyApplicationPage {

	my ($template) = HTML::Template->new(
		filename => "$htmldir/proxy-application-form.html" );

	$template->param( "cgi_url", $cgi_url );

	# We only want Halls on the building list
	my ($building_selector) = BuildBuildingSelector( 2 );
	$template->param( "building_selector", $building_selector );

	my ($semester_selector) = BuildSemesterSelector("", 0);
	$template->param( "semester_selector", $semester_selector );

	EmitPage( "Halls Networking Applicant",
		"Form to record details of a student wanting network access in their Hall room",
		$template->output() );
}


########################################################################
#
# Trim off leading and trailing white space from a variable
#
########################################################################
sub Trim {
	foreach (@_) {
		next if (!defined($_));
		s/^\s*//;
		s/\s*$//;
	}
}

########################################################################
#
# Fix up upper and lower-case to a canonical form
#
########################################################################
sub FixCase {
	foreach (@_) {
		next if (!defined($_));
		# Force all to lower case then first char to upper
		$_ = "\u\L$_";
	}
}

########################################################################
#
# Check that supplied value contains alphanum chars only
#
########################################################################
sub OKAlphaNum {
	my ($value) = @_;

	if ($value =~ /[^a-zA-Z0-9]/) {
		# Bad char found
		return 0;
	}
	else {
		# OK
		return 1;
	}
}


########################################################################
#
# Check that supplied value contains numbers and spaces only
#
########################################################################
sub OKNumSpace {
	my ($value) = @_;

	if ($value =~ /[^0-9 ]/) {
		# Bad char found
		return 0;
	}
	else {
		# OK
		return 1;
	}
}

########################################################################
#
# Check that supplied value contains alphanum and punctuation chars only
#
########################################################################
sub OKAlphaNumPunct {
	my ($value) = @_;

	# Dont put HTML special chars in this string:
	if ($value =~ /[^-a-zA-Z0-9.,\/: ]/) {
		# Bad char found
		return 0;
	}
	else {
		# OK
		return 1;
	}
}

########################################################################
#
# Get card details from the supplied query and return cleaned data.
# Usage:
#	my ($err, $cardnumber, $cardname, $cardexpiry, $cardissue) =
#		GetCardData( $query );
#
# If all OK, $err will be zero. Otherwise it will contain an error
# message and the other values will be undefined
#
########################################################################
sub GetCardData {
	my ($query) = @_;

	my ($cardnumber, $cardname, $cardexpiry, $cardissue);

	$cardnumber = $query->param("cardnumber");
	Trim($cardnumber);
	if (!defined($cardnumber) || !OKNumSpace($cardnumber) ||
		!($cardnumber =~ /[0-9]/) ) {
		return( "The card number does not look right. It should contain only numbers and spaces",
		undef, undef, undef, undef);
	}
	# Check credit-card number checksum
	if (!validate($cardnumber)) {
		return( "The card number is not valid. It should look something like \"5224&nbsp;0123&nbsp;4567&nbsp;8901\" Please check and try again.",
		undef, undef, undef, undef);
	}

	$cardname = $query->param("cardname");
	Trim($cardname);
	if (!defined($cardname) || !OKAlphaNumPunct($cardname) ||
		!($cardname =~ /[a-zA-Z]/) ) {
		return( "The account name does not look right. It should contain only letters, numbers and spaces",
		undef, undef, undef, undef);
	}

	$cardexpiry = $query->param("cardexpiry");
	Trim($cardexpiry);
	if (!defined($cardexpiry) || !OKAlphaNumPunct($cardexpiry) ||
		!($cardexpiry =~ /[0-9]/) ) {
		return( "The expiry date does not look right. It should contain only numbers and '/'",
		undef, undef, undef, undef);
	}

	$cardissue = $query->param("cardissue");
	Trim($cardissue);
	if (defined($cardissue) && !OKNumSpace($cardissue)) {
		return( "The issue number does not look right. It should contain only numbers.",
		undef, undef, undef, undef);
	}
	$cardissue = "" if (!defined($cardissue));

	# All OK!
	return (0, $cardnumber, $cardname, $cardexpiry, $cardissue);
}

########################################################################
#
# Emit an "Apply for Halls" page
# - actually handles several similar pages
#
# Note that this also validates a lot of form data
#
########################################################################
sub EmitApplyForHallsPaymentPage {

	my ($stage) = @_;

	my ($template, $pagetitle, $pagedesc, $applicant_username);
	my ($mail_template);

	my ($is_proxy) = $query->param("proxy");
	if (defined($is_proxy) && ($is_proxy eq "yes")) {
		$is_proxy = 1;
	}
	else {
		$is_proxy = 0;
	}

	# Proxy applications can only be handled by billing staff
	if ($is_proxy && !Permitted( $loggedinuser, "billing" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for proxy application" );
		exit 1;
	}

	if ($is_proxy) {
		my ($user) = $query->param("user");
		Trim($user);

		if (!defined($user)) {
			EmitError( "No user specified" );
			return 1;
		}

		if ($user =~ /[^a-zA-Z0-9]/) {
			EmitError( "Bad characters in username" );
			return 1;
		}

		$applicant_username = $user;
	}
	else {
		$applicant_username = $loggedinuser;
	}

	my ($fullname) = FetchFullName( $applicant_username, 0 );
	if (!$fullname) {
		EmitError( "No entry found for user '$applicant_username'" );
		return 1;
	}

	# We must have a valid semester
	my($semester) = $query->param("semester");
	Trim($semester);
	if (!ValidateInputValue("semesters", "sem_id", $semester)) {
		EmitError( "Bad URL or form: no valid semester was specified" );
		exit 1;
	}

	# Get details about that semester
	my (@sem_details) = FetchSemester( $semester );
	if (!@sem_details) {
		EmitError( "Internal error: missing details for semester id $semester" );
		exit 1;
	}
	my ($start_date, $end_date) = ($sem_details[1], $sem_details[2]);

	# We can now calculate the total cost of the request.
	# Do this early so we can avoid asking for nil payments
	my ($totalcost) = sprintf "%6.2f", $sem_details[3]+$sem_details[4];


	# We create a mail template for both stage 2 and stage 3
	# as it is simpler to do so. We will only use it in stage 3.
	$mail_template = HTML::Template->new(
		filename => "$htmldir/mail-application-complete.txt" );

	if ($stage == 2) {
		if ($is_proxy) {
		    if ($totalcost == 0) {
			$template = HTML::Template->new(
				filename => "$htmldir/proxy-nocharge-form.html" );
		    }
		    else {
			$template = HTML::Template->new(
				filename => "$htmldir/proxy-payment-form.html" );
		    }
		}
		else {
		    if ($totalcost == 0) {
			$template = HTML::Template->new(
				filename => "$htmldir/halls-nocharge-form.html" );
		    }
		    else {
			$template = HTML::Template->new(
				filename => "$htmldir/halls-payment-form.html" );
		    }
		}
		$pagetitle = "Confirm details of network service";
		$pagedesc = "Confirm details for network service in Halls of Residence";
	}
	elsif ($stage == 3) {
		$template = HTML::Template->new(
			filename => "$htmldir/halls-application-complete.html" );
		$pagetitle = "Application Complete";
		$pagedesc = "Completed application for network service in Halls of Residence";
	}
	else {
		EmitError( "Internal error: unknown payment stage" );
		exit 1;
	}

	$template->param( "user", $applicant_username );
	$mail_template->param( "user", $applicant_username );
	$template->param( "fullname", $fullname );
	$mail_template->param( "fullname", $fullname );
	$template->param( "cgi_url", $cgi_url ) if ($stage == 2);
	$template->param( "semester", $semester ) if ($stage == 2);

	# For proxy applications we want a quick way back to the start of the process
	if ($is_proxy && ($stage == 3)) {
		my ($arenafrag) = BuildArenaFrag();
		$template->param( "arenamodefrag", $arenafrag );
	}

	# We must have a valid building
	my($building) = $query->param("building");
	Trim($building);
	if (!ValidateInputValue("buildings", "building_id", $building)) {
		EmitError( "You must select the building where you want a connection." );
		exit 1;
	}
	$template->param( "building", $building ) if ($stage == 2);
	my ($building_name) = FetchBuildingName( $building );
	if (!$building_name) {
		# This should not be possible as we verified the building above!
		EmitError( "Internal error: missing building description" );
		exit 1;
	}
	$template->param( "building_name", $building_name );
	$mail_template->param( "building_name", $building_name );

	# There may be a block number
	my($block) = $query->param("block");
	Trim($block);
	if (defined($block) && !OKAlphaNum($block)) {
		EmitError( "A bad value was specified for 'block'" );
		exit 1;
	}
	$block = "" if !defined($block);
	FixCase($block);
	$template->param( "block", $block );
	$mail_template->param( "block", $block );

	# There may be a room number
	my($room) = $query->param("room");
	Trim($room);
	if (defined($room) && !OKAlphaNumPunct($room)) {
		EmitError( "A bad value was specified for 'room'" );
		exit 1;
	}
	$room = "" if !defined($room);
	$template->param( "room", $room );
	$mail_template->param( "room", $room );

	# There may be an outlet number
	my($outlet) = $query->param("outlet");
	Trim($outlet);
	if (defined($outlet) && !OKAlphaNumPunct($outlet)) {
		EmitError( "A bad value was specified for 'outlet'" );
		exit 1;
	}
	$outlet = "" if !defined($outlet);

	# Sanity checks
	if ($room eq "") {
		EmitError( "You must fill in the room number." );
		exit 1;
	}

	if (($block eq "") && HasBlocks($building)) {
		EmitError( "$building_name has flats or blocks but you have not filled in the flat/block number" );
		exit 1;
	}

	if (($block ne "") && !HasBlocks($building)) {
		EmitError( "You have filled in a flat/block number but $building_name does not have flats or blocks." );
		exit 1;
	}


	# Avoid students applying for service many times over
	if (($mode eq "halls") && (!$TestMode) && (HallsRequestExists($applicant_username, $start_date) > 0)) {
		EmitError( "You have already applied for a Halls network service starting on $start_date" );
		exit 1;
	}

	#-# We would like to validate the outlet number if one is supplied,
	#-# but cannot do so until our database of outlets is complete

	#-# Temporary requirement: we have no outlet info for Fleming Hall
	#-# so all requests must include outlet number until we get it.
	#-# if (($outlet eq "") && ($building_name =~ /^Fleming/)) {
	#-# 	EmitError( "Please fill in the number of the network outlet in your room" );
	#-# 	exit 1;
	#-# }


	# If no outlet number was supplied, try to fill it in from the database
	# using the lowest-numbered outlet in the room
	if ($outlet eq "") {
		my ($outlet_list) = FetchOutlets( $building, $block, $room );
		if (defined($outlet_list->[0][1])) {
			$outlet = $outlet_list->[0][1];
		}
	}

	$template->param( "outlet", $outlet );
	$mail_template->param( "outlet", $outlet );

	$template->param( "start_date", HumaniseDate($start_date) );
	$mail_template->param( "start_date", HumaniseDate($start_date) );
	$template->param( "end_date", HumaniseDate($end_date) );
	$mail_template->param( "end_date", HumaniseDate($end_date) );
	$template->param( "conn_fee", $sem_details[3] );
	$mail_template->param( "conn_fee", $sem_details[3] );
	$template->param( "sem_fee", $sem_details[4] );
	$mail_template->param( "sem_fee", $sem_details[4] );
	$template->param( "totalcost", $totalcost );
	$mail_template->param( "totalcost", $totalcost );


	####################################################################
	# Everything from here on is stage 3 only, so stage 2 returns now
	####################################################################
	if ($stage == 2) {
		EmitPage( $pagetitle, $pagedesc, $template->output() );
		return;
	}


	# We must have a valid payment method
	# Each method has a set of instructions associated with it
	my($payment_method) = $query->param("payment_method");
	Trim($payment_method);
	$payment_method = "" if (!defined($payment_method));
	my($payfrag);
	if ($payment_method eq "card") {
		$payfrag = "how-to-pay-done-frag.html";
	}
	elsif ($payment_method eq "post") {
		$payfrag = "how-to-pay-by-post-frag.html";
	}
	elsif ($payment_method eq "cash") {
		$payfrag = "how-to-pay-by-cash-frag.html";
	}
	elsif ($payment_method eq "paid") {
		$payfrag = "how-to-pay-paid-frag.html";
	}
	elsif (($payment_method eq "nocharge") && ($totalcost == 0)) {
		$payfrag = "how-to-pay-nocharge-frag.html";
	}
	else {
		EmitError( "Bad URL or form: no valid payment_method was specified" );
		exit 1;
	}

	# How much money are we being offered?
	my ($payment) = $query->param("payment");
	Trim($payment);
	if (!defined($payment)) {
		$payment = 0;
	}
	if ($payment =~ /[^0-9.]/) {
		EmitError( "Bad characters in payment" );
		return 1;
	}
	# Have we taken the right amount of money?
	if ($is_proxy && ($payment_method eq "paid") && ($payment != $totalcost)) {
		EmitError( "Payment is wrong: &pound;$totalcost expected" );
		exit 1;
	}

	my($tem2) = HTML::Template->new( filename => "$htmldir/$payfrag" );
	$template->param("howtopayfrag", $tem2->output());

	# If payment will be by online card details we must get them and check
	# that they are reasonable
	my ($cardnumber, $cardname, $cardexpiry, $cardissue);
	if ($payment_method eq "card") {
		my ($err);

		($err, $cardnumber, $cardname, $cardexpiry, $cardissue) =
			GetCardData( $query );

		if ($err) {
			EmitError( $err );
			exit 1;
		}
	}

	# The result page is now set up.
	# It is time to record the transaction in the database!

	# Add the user to the database if not already there
	if (!FetchFullName($applicant_username, 1)) {
		if (!AddUser($applicant_username)) {
			EmitError( "Internal error: cannot add user $applicant_username to the database." );
			exit 1;
		}
	}

	# Add the card details to the database
	my ($bill_id) = 0;
	if ($payment_method eq "card") {
		$bill_id = AddCardDetails( $cardname, $cardnumber, $cardexpiry, $cardissue );
		if (!$bill_id) {
			EmitError( "Internal error: AddCardDetails failed $ConnDB::errstr" );
			exit 1;
		}
	}

	my ($req_comment) = "";
	if ($is_proxy) {
		$req_comment = "Proxy application processed by $loggedinuser";
	}

	# Now add a service request
	my ($srvreqid) = AddServiceRequest(
		$applicant_username, $CONST_halls_service, $totalcost, $bill_id, $req_comment );
	if (!$srvreqid) {
		EmitError( "Internal error: $ConnDB::errstr" );
		Log( "err", "AddServiceRequest failed: $ConnDB::errstr" );
		# *** Should delete the billing details record here
		exit 1;
	}
	$template->param("servreqid", $srvreqid);
	$mail_template->param("servreqid", $srvreqid);

	# Now add the details of the service required
	if (!AddServiceRequestItem(
		$srvreqid, $building, $block, $room, $outlet, "10M",
		$start_date, $end_date, $CONST_srv_state_awaiting_payment,
		$totalcost, "" )) {
			EmitError( "Internal error: $ConnDB::errstr" );
			Log( "err", "AddServiceRequestItem failed: $ConnDB::errstr" );
			# *** Should delete service request and billing details records here
			exit 1;
	}

	# If cost is zero, accept zero payment here
	if ($totalcost == 0) {
		$payment = 0;
	}

	# Proxy applications may be pre-paid
	# If so, process the payment now.
	if (($totalcost == 0) || ($is_proxy && ($payment_method eq "paid"))) {
		if (!ProcessPrePayment( $srvreqid, $payment )) {
			EmitError( "Service request has been registered but ProcessPrePayment failed: $ConnDB::errstr" );
			Log( "err", "ProcessPrePayment failed: $ConnDB::errstr" );
			# *** Should consider deleting the service request here
			exit 1;
		}
	}

	# Get the timestamp that was applied to this request
	my ($service_requests) = FetchRequests( "%", $srvreqid );
	my ($timestamp) = "MISSING";
	if ($service_requests) {
		$timestamp = $service_requests->[0][3];
	}
	$template->param( "timestamp", HumaniseDate($timestamp) );
	$mail_template->param( "timestamp", HumaniseDate($timestamp) );

	# Dont talk about payments when they are not needed
	if ($totalcost == 0) {
		$mail_template->param( "connection_note",
			"Your connection will be made as soon as possible." );
	}
	else {
		$mail_template->param( "connection_note",
		"When your payment has been processed the connection will be made." );
	}

	# Send all mail to one place while testing
	my ($recipient) = $applicant_username;
	if ($TestMode) {
		$recipient = $TestRecipient;
	}

	# Inform user by mail
	SendMail( $recipient, 'Halls Network Service', $mail_template->output() );

	# Inform user by Web
	EmitPage( $pagetitle, $pagedesc, $template->output() );
}

########################################################################
# Emit a "payment for renewal" page
########################################################################
sub EmitHallsRenewalPaymentPage {

	my ($stage) = @_;

	my ($pagetitle, $pagedesc, $applicant_username);
	my ($mail_template);

	# Get ID for item to be renewed
	my ($item_id) = $query->param("item_id");
	Trim($item_id);
	if (!defined($item_id) || ($item_id < 0)) {
		EmitError( "You must specify an item to renew" );
		exit 1;
	}

	# Get the full details for this service_request_item
	my (
		$service_request_id,
		$building_id,
		$building_name,
		$block_number,
		$room_number,
		$outlet_number,
		$port_type,
		$start_date,
		$end_date,
		$service_state_id,
		$service_state_description,
		$comment,
		$username
	) = FetchRequestItem( $item_id );

	if (!defined($username)) {
		EmitError( "Unknown item_id specified" );
		exit 1;
	}

	if ($service_state_id != 3) {
		EmitError( "That connection is `$service_state_description'. Only `connected' connections can be extended" );
		exit 1;
	}

	# Proxy applications can only be handled by billing staff
	my ($is_proxy) = ($username ne $loggedinuser);

	if ($is_proxy && !Permitted( $loggedinuser, "billing" )) {
		EmitError( "You cannot renew another user's connection" );
		Log( "error", "$applicant_username trying to renew item $item_id which is owned by $username" );
		exit 1;
	}

	if ($is_proxy) {
		# Take username from item in database
		$applicant_username = $username;
	}
	else {
		$applicant_username = $loggedinuser;
	}

	my ($fullname) = FetchFullName( $applicant_username, 1 );
	if (!$fullname) {
		EmitError( "No existing connection found for user '$applicant_username'" );
		return 1;
	}


	# Get ID of semester being requested
	my ($sem_id) = $query->param("semester");
	Trim($sem_id);
	if (!defined($sem_id) || ($sem_id < 0)) {
		EmitError( "You must specify a semester for this renewal" );
		exit 1;
	}

	#### We must check that semester follows on from a
	#### currently booked and paid-for time

	# Get list of allowable renewal semesters 
	my ($semester_list) = FetchRenewalSemesterList($end_date,
					$CONST_semester_days_tolerance);

	# Walk through the list to find the one selected
	my ($sem);
	my ($sem_end_date, $sem_fee, $sem_descr);
	for ($sem=0;defined($semester_list->[$sem]);$sem++) {
		# Is this the right sem_id?
		next if ($semester_list->[$sem][0] != $sem_id);

		# Yes, so record data
		$sem_end_date = $semester_list->[$sem][2];
		$sem_fee = $semester_list->[$sem][4];
		$sem_descr = $semester_list->[$sem][5];
	}

	if (!defined($sem_end_date)) {
		EmitError( "Unknown or invalid semester specified" );
		exit 1;
	}

	# OK - the request is valid.
	# From here on things diverge:
	#	Stage 1 was the account page that got us to here.
	#	In stage 2 we emit a payment options page
	#	In stage 3 we process the result of that and emit
	#	a confirmation page

	# Prepare the page template
	my ($template);
	if ($stage == 2) {
		if ($sem_fee == 0) {
		    $template = HTML::Template->new(
			filename => "$htmldir/renew-nocharge-form.html" );
		}
		else {
		    $template = HTML::Template->new(
			filename => "$htmldir/renew-payment-form.html" );
		}
		$template->param( "item_id", $item_id );
		$template->param( "sem_id", $sem_id );
	}
	if ($stage == 3) {
		$template = HTML::Template->new(
			filename => "$htmldir/renew-complete.html" );
		$template->param( "servreqid", $service_request_id );
	}

	$template->param( "cgi_url", $cgi_url );
	$template->param( "building_name", $building_name );
	$template->param( "block", $block_number );
	$template->param( "room", $room_number );
	$template->param( "outlet", $outlet_number );
	$template->param( "end_date", HumaniseDate($end_date) );
	$template->param( "new_end_date", HumaniseDate($sem_end_date) );
	$template->param( "user", $username );
	$template->param( "fullname", $fullname );
	$template->param( "sem_fee", $sem_fee );
	$template->param( "totalcost", $sem_fee );


	if ($stage == 2) {
		# Just emit the page and return
		EmitPage( "Extension of halls network connection: $applicant_username",
			"Extension of halls network connection for $applicant_username",
			$template->output()
		);

		Log( "info", "renewal payment options page for $applicant_username" );
		return;
	}

	# All stage 3 from here on

	# We must have a valid payment method
	# Each method has a set of instructions associated with it
	my($payment_method) = $query->param("payment_method");
	Trim($payment_method);
	$payment_method = "" if (!defined($payment_method));
	my($payfrag);
	if ($payment_method eq "card") {
		$payfrag = "how-to-pay-done-frag.html";
	}
	elsif ($payment_method eq "post") {
		$payfrag = "how-to-pay-by-post-frag.html";
	}
	elsif ($payment_method eq "cash") {
		$payfrag = "how-to-pay-by-cash-frag.html";
	}
	elsif ($payment_method eq "paid") {
		$payfrag = "how-to-pay-paid-frag.html";
	}
	elsif (($payment_method eq "nocharge") && ($sem_fee == 0)) {
		$payfrag = "renew-done-nocharge-frag.html";
	}
	else {
		EmitError( "Bad URL or form: no valid payment_method was specified" );
		exit 1;
	}

	my($tem2) = HTML::Template->new( filename => "$htmldir/$payfrag" );
	$template->param("howtopayfrag", $tem2->output());

	# If payment will be by online card details we must get them and check
	# that they are reasonable
	my ($cardnumber, $cardname, $cardexpiry, $cardissue);
	if ($payment_method eq "card") {
		my ($err);

		($err, $cardnumber, $cardname, $cardexpiry, $cardissue) =
			GetCardData( $query );

		if ($err) {
			EmitError( $err );
			exit 1;
		}
	}

	#######################################################
	# We now have all the data we need so start the updates
	#######################################################

	# Add the card details to the database
	my ($bill_id) = 0;
	if ($payment_method eq "card") {
		$bill_id = AddCardDetails( $cardname, $cardnumber, $cardexpiry, $cardissue );
		if (!$bill_id) {
			EmitError( "Internal error: AddCardDetails failed $ConnDB::errstr" );
			exit 1;
		}
	}

	# Update the database
	ApplyForExtension( $service_request_id, $item_id,
				$sem_end_date, $sem_fee, $bill_id );

	# If no charge, process the null payment now
	if ($sem_fee == 0) {
		if (!ProcessPrePayment( $service_request_id, 0 )) {
			EmitError( "Service request has been registered but ProcessPrePayment failed: $ConnDB::errstr" );
			Log( "err", "ProcessPrePayment failed in nocharge renew: $ConnDB::errstr" );
			# *** Should consider deleting the service request here
			exit 1;
		}
	}

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	EmitPage( "Application Complete: extension for $applicant_username",
		"Application Complete: extension for $applicant_username",
		$template->output()
	);

	Log( "info", "renewal application complete for $applicant_username" );
}

########################################################################
# Build a cancel-request fragment for a service_request_item
########################################################################
sub BuildCancelFrag {
	my ( $service_request_item_id) = @_;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/cancel-item-frag.html" );

	$template->param( "cgi_url", $cgi_url );
	$template->param( "item_id", $service_request_item_id );

	return $template->output;
}

########################################################################
# Build a renew-request fragment for a service_request_item
########################################################################
sub BuildRenewFrag {
	my (
		$service_request_item_id,
		$current_end_date
	) = @_;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/renew-item-frag.html" );

	my ($semester_selector) = BuildSemesterSelector(
					$current_end_date,
					$CONST_semester_days_tolerance
				);
	$template->param( "cgi_url", $cgi_url );
	$template->param( "item_id", $service_request_item_id );
	$template->param( "semester_selector", $semester_selector );

	return $template->output;
}


########################################################################
# Build a fragment to provide quick navigation links in proxy mode
########################################################################
sub BuildArenaFrag {

	my ($template) = HTML::Template->new(
		filename => "$htmldir/arena-mode-trailer-frag.html" );

	$template->param( "cgi_url", $cgi_url );

	return $template->output;
}


########################################################################
# Utility procedure to build an array of hashes to describe line-items
########################################################################

sub BuildLineItemsArray {
	my (
		$srv_req_id,
		$add_cancel_frag,
		$add_renew_frag,
		$req_date,
		$paid_date
	) = @_;
	# Passing req_date and paid_date is a bit messy but it saves
	# an extra DB lookup.

	my ($line_items) = FetchRequestItemsByRequestID( $srv_req_id );
	if (!$line_items) {
		EmitError( "FetchRequestItemsByRequestID failed $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my (@item_hash_array);
	for ($item=0;defined($line_items->[$item]);$item++) {
		# A bit of tidying up
		my ($clean_end_date) = $line_items->[$item][17];
		my ($clean_con_building) = $line_items->[$item][1];
		if (!defined($clean_end_date) ||
				($clean_end_date eq "0000-00-00 00:00:00")) {
			# Only show date if it is non-zero
			$clean_end_date = "";
		}
		if (!defined($line_items->[$item][10])) {
			# Only show building if connection has been made
			$clean_con_building = "";
		}

		my ($connectionstate_id) = $line_items->[$item][21];

		# Provide a cancel request button if appropriate
		my ($cancelfrag) = "";
		if ($add_cancel_frag && ($connectionstate_id =~ /^(1|2|3|7)$/)) {
			$cancelfrag = BuildCancelFrag( $line_items->[$item][20] );
		}

		# Provide a renew request option if appropriate
		my ($renewfrag) = "";
		if ($add_renew_frag && ($connectionstate_id =~ /^(3)$/)) {
			$renewfrag = BuildRenewFrag(
				# request_item_id
				$line_items->[$item][20],
				# end_date
				HumaniseDate($line_items->[$item][6])
			);
		}

		$item_hash_array[$item] = {
			"building" => $line_items->[$item][1],
			"block_number" => $line_items->[$item][2],
			"room" => $line_items->[$item][3],
			"outlet_number" => $line_items->[$item][4],
			"req_date" => HumaniseDate($req_date),
			"paid_date" => HumaniseDate($paid_date),
			"startdate" => HumaniseDate($line_items->[$item][5]),
			"enddate" => HumaniseDate($line_items->[$item][6]),
			"connectionstate" => $line_items->[$item][7],
			"itemcomment" => $line_items->[$item][8],
			"con_building" => $clean_con_building,
			"con_block_number" => $line_items->[$item][10],
			"con_room" => $line_items->[$item][11],
			"con_outlet_number" => $line_items->[$item][12],
			"con_device_name" => $line_items->[$item][14],
			"con_device_port" => $line_items->[$item][15],
			"con_startdate" =>
				HumaniseDate($line_items->[$item][16]),
			"con_enddate" => HumaniseDate($clean_end_date),
			"con_conby" => $line_items->[$item][18],
			"con_disby" => $line_items->[$item][19],
			"cancelfrag" => $cancelfrag,
			"renewfrag" => $renewfrag
		}
	}

	return @item_hash_array;

}

########################################################################
# Build a billing-details fragment for an output page
########################################################################
sub BuildBillingDetailsFrag
{
	my ($bill_det_id) = @_;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/billing-frag.html" );

	# Billing details
	my ($cardnum, $cardexpire, $cardissue, $cardname,
		$internal_charge_code);
	
	if ($bill_det_id > 0) {
		# We have billing details
		my (@bill_dets) = FetchBillingDetails( $bill_det_id );
		if (!@bill_dets) {
			EmitError(
				"FetchBillingDetails failed $ConnDB::errstr" );
			return 1;
		}
		$cardname = $bill_dets[0];
		$cardnum = $bill_dets[1];
		$cardexpire = $bill_dets[2];
		$cardissue = $bill_dets[3];
		$internal_charge_code = $bill_dets[4];

		# Leave a space char in unused elements to force browser
		# to render the background colour
		$cardname = "&nbsp;" if ($cardname eq "");
		$cardnum = "&nbsp;" if ($cardnum eq "");
		$cardexpire = "&nbsp;" if ($cardexpire eq "");
		$cardissue = "&nbsp;" if ($cardissue eq "");
	}
	else {
		return "";
	}


	$template->param( "cardname", $cardname );
	$template->param( "cardnum", $cardnum );
	$template->param( "cardexpire", $cardexpire );
	$template->param( "cardissue", $cardissue );
	$template->param( "internal_charge_code", $internal_charge_code );

	return $template->output;
}

########################################################################
# Build an accept-payment button fragment for an output page
########################################################################
sub BuildAcceptPaymentFrag {
	my ( $user, $servreqid ) = @_;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/accept-payment-frag.html" );

	$template->param( "user", $user );
	$template->param( "servreqid", $servreqid );
	$template->param( "paymenturl", $cgi_url );

	return $template->output;
}


########################################################################
# Emit a User Account page
########################################################################
sub EmitUserAccount {

	my ($template) = HTML::Template->new(
		filename => "$htmldir/user-account.html" );

	my ($user) = $query->param("user");
	Trim($user);

	my ($useloggedinuser) = $query->param("useloggedinuser");
	Trim($useloggedinuser);
	if (defined($useloggedinuser) && ($useloggedinuser eq "yes")) {
		# Simplify access to user's own account
		$user = $loggedinuser
	}

	if (!defined($user)) {
		EmitError( "No user specified" );
		return 1;
	}

	if ($user =~ /[^a-zA-Z0-9]/) {
		EmitError( "Bad characters in username" );
		return 1;
	}

	my ($fullname) = FetchFullName( $user, 1 );
	if (!$fullname) {
		EmitError( "No networking requests found for user '$user'" );
		return 1;
	}

	my ($service_requests) = FetchRequests( $user, "%" );
	if (!$service_requests) {
		EmitError( "FetchRequests failed $ConnDB::errstr" );
		return 1;
	}

	my ($has_billing) = Permitted( $loggedinuser, "billing" );
	my ($has_connect) = Permitted( $loggedinuser, "connect" );

	# Unpick the array of arrays and build an array of hashes
	my ($req);
	my (@req_hash_array);
	for ($req=0;defined($service_requests->[$req]);$req++) {
		# We dont really want the time part here
		my ($clean_req_date) = $service_requests->[$req][3];
		$clean_req_date =~ s/ .*//;
		if ($clean_req_date eq "0000-00-00") {
			# Only show date if it is non-zero
			$clean_req_date = "";
		}

		my ($clean_paid_date) = $service_requests->[$req][8];
		$clean_paid_date =~ s/ .*//;
		if ($clean_paid_date eq "0000-00-00") {
			# Only show date if it is non-zero
			$clean_paid_date = "";
		}

		# Each request can have several line-items
		# so we build an array to describe them
		my (@line_items_array) =
			BuildLineItemsArray( $service_requests->[$req][0],
				($has_billing || $has_connect || ($user eq $loggedinuser)),
				($has_billing || ($user eq $loggedinuser)),
				$clean_req_date,
				$clean_paid_date );

		my ($accept_payment_frag) = "";
		my ($billing_details_frag) = "";

		# Billing details
		if ($has_billing) {
			$billing_details_frag = BuildBillingDetailsFrag (
					$service_requests->[$req][6]
			);

			# If charge > amount_paid then provide an Accept Payment button
			if ($service_requests->[$req][4] > $service_requests->[$req][5]) {
				# Accept-payment button
				$accept_payment_frag = BuildAcceptPaymentFrag(
					$user, $service_requests->[$req][0]
				);
			}
		}

		# Now build the hash describing the main service request
		$req_hash_array[$req] = {
			"servreqid" => $service_requests->[$req][0],
			"requestuser" => $service_requests->[$req][1],
			"requestdate" => HumaniseDate($service_requests->[$req][3]),
			"charge" => $service_requests->[$req][4],
			"amountpaid" => $service_requests->[$req][5],
			"billing_details" => $billing_details_frag,
			"reqcomment" => $service_requests->[$req][7],
			"lineitems" => \@line_items_array,
			"accept_payment_frag" => $accept_payment_frag
		}
	}


	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	$template->param( "requests", \@req_hash_array );
	$template->param( "user", $user );
	$template->param( "fullname", $fullname );

	EmitPage( "Halls network account: $user",
		"Halls network account for $user",
		$template->output()
	);

	Log( "info", "account page for $user" );
}

########################################################################
# Emit a Payment List page
########################################################################
sub EmitPaymentList {
	if (!Permitted( $loggedinuser, "billing" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for payments list" );
		exit 1;
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/payment-list.html" );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	my ($list_items) = FetchPendingCardPayments;
	if (!$list_items) {
		EmitError( "FetchPendingCardPayments failed $ConnDB::errstr" );
		Log( "err", "FetchPendingCardPayments failed $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my (@item_hash_array);
	for ($item=0;defined($list_items->[$item]);$item++) {
		$item_hash_array[$item] = {
			"username" => $list_items->[$item][1],
			"fullname" => $list_items->[$item][2],
			"cgiurl" => $cgi_url
		}
	}
	$template->param( "cardusers", \@item_hash_array );

	# Now do the same for cash/cheque users
	my($cash_list_items) = FetchPendingCashPayments;
	if (!$cash_list_items) {
		EmitError( "FetchPendingCashPayments failed $ConnDB::errstr" );
		Log( "err", "FetchPendingCashPayments failed $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my (@cash_item_hash_array);
	for ($item=0;defined($cash_list_items->[$item]);$item++) {
		$cash_item_hash_array[$item] = {
			"username" => $cash_list_items->[$item][1],
			"fullname" => $cash_list_items->[$item][2],
			"cgiurl" => $cgi_url
		}
	}
	$template->param( "cashusers", \@cash_item_hash_array );

	EmitPage( "Halls Network Payment Processing",
		"Halls Network Payment Processing",
		$template->output()
	);

	Log( "info", "payments page" );
}

########################################################################
# Accept a payment
########################################################################
sub AcceptPayment {

	if (!Permitted( $loggedinuser, "billing" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for accept payment" );
		exit 1;
	}

	# Get the service request ID
	my ($servreqid) = $query->param("servreqid");
	Trim($servreqid);
	if (!defined($servreqid) || ($servreqid < 0)) {
		EmitError( "No service request ID was specified" );
		exit 1;
	}
	if ($servreqid =~ /[^0-9]/) {
		EmitError( "Bad characters in servreqid" );
		return 1;
	}


	# How much money are we being offered?
	my ($payment) = $query->param("payment");
	Trim($payment);
	if (!defined($payment) || ($payment < 0)) {
		EmitError( "Attempt to accept null payment" );
		exit 1;
	}
	if ($payment =~ /[^0-9.]/) {
		EmitError( "Bad characters in payment" );
		return 1;
	}

	# Work out the username for that request ID
	my ($service_requests) = FetchRequests( "%", $servreqid );
	if (!$service_requests) {
		EmitError( "FetchRequests failed: $ConnDB::errstr" );
		Log( "err", "FetchRequests failed: $ConnDB::errstr" );
		return 1;
	}
	my ($user) = $service_requests->[0][1];

	# Request looks OK - try to process it
	if (!ProcessPrePayment( $servreqid, $payment )) {
		EmitError( "ProcessPrePayment failed: $ConnDB::errstr" );
		Log( "err", "ProcessPrePayment failed: $ConnDB::errstr" );
		exit 1;
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/payment-accepted.html" );

	my ($mail_template) = HTML::Template->new(
		filename => "$htmldir/mail-payment-accepted.txt" );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );
	$mail_template->param( "date", HumaniseDate($date) );


	$template->param( "user", $user );
	# $mail_template->param( "user", $user );
	$template->param( "servreqid", $servreqid );
	# $mail_template->param( "servreqid", $servreqid );
	$template->param( "payment", $payment );
	$mail_template->param( "payment", $payment );
	$template->param( "cgi_url", $cgi_url );

	my ($recipient) = $user;
	if ($TestMode) {
		$recipient = $TestRecipient;
	}

	# Inform user by mail (but only if they actually paid something!)
	if ($payment != 0) {
		SendMail( $recipient,
			'Halls Network Payment', $mail_template->output() );
	}

	EmitPage( "Payment Accepted", "Payment has been accepted from $user", $template->output() );
}


########################################################################
# BuildConnectRow - one row of widgets for the connection list
########################################################################
sub BuildConnectRow {
	my ( $req_item_id, $room_number, $outlet_number,
		$device_selector, $user, $fullname ) = @_;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/connection-list-con-frag.html" );

	$template->param( "req_id", $req_item_id );
	$template->param( "room", $room_number );
	$template->param( "outlet_number", $outlet_number );
	$template->param( "cgiurl", $cgi_url );
	$template->param( "device_selector", $device_selector );
	$template->param( "username", $user );
	$template->param( "fullname", $fullname );

	return $template->output();
}


########################################################################
# BuildDisconnectRow - one row of widgets for the connection list
########################################################################
sub BuildDisconnectRow {
	my ( $req_item_id, $room_number, $outlet_number,
		$device, $port, $user, $fullname ) = @_;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/connection-list-dis-frag.html" );

	$template->param( "req_id", $req_item_id );
	$template->param( "room", $room_number );
	$template->param( "outlet_number", $outlet_number );
	$template->param( "device", $device );
	$template->param( "port", $port );
	$template->param( "cgiurl", $cgi_url );
	$template->param( "username", $user );
	$template->param( "fullname", $fullname );

	return $template->output();
}

########################################################################
# Emit a Connection Job List page
########################################################################
sub EmitConnectionList {

	# Allow requester to limit display by site
	my ($campus_spec) = $query->param("site_id");
	Trim($campus_spec);
	$campus_spec = "%" if (!defined($campus_spec) || ($campus_spec eq ""));

	# Allow requester to limit display by building
	my ($building_spec) = $query->param("building");
	Trim($building_spec);
	$building_spec = "%" if (!defined($building_spec) || ($building_spec eq ""));


	my ($template) = HTML::Template->new(
		filename => "$htmldir/connection-list.html" );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );


	my ($list_items) = FetchPendingConnections($campus_spec, $building_spec);
	if (!$list_items) {
		EmitError( "FetchPendingConnections failed $ConnDB::errstr" );
		Log( "err", "FetchPendingConnections failed: $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my ($bldg_ser) = 0;
	my (@bldg_hash_array);
	my ($job_hash_array);
	my ($job_ser);
	my ($last_bldg_id) = "";
	for ($item=0;defined($list_items->[$item]);$item++) {
		my ($req_item_id, $req_id, $user, $fullname, $req_date, $req_comment,
			$building_id, $building,
			$site_name, $service_state_id,
			$block_number, $room_number, $outlet_number,
			$start_date, $end_date, $item_comment) =
				@{$list_items->[$item]};

		# If this is a new building then we start a new array
		if ($last_bldg_id ne $building) {
			# Create a job array for the new building
			$job_ser = 0;
			# This weird construct gets us a reference to a new array:
			$job_hash_array = \@{[]};
			$bldg_hash_array[$bldg_ser] = {
				"building" => $building,
				"items" => $job_hash_array
			};
			$bldg_ser++;
			$last_bldg_id = $building;
		}

		my ($job_row);

		# Are we connecting or disconnecting here?
		if ($service_state_id == $CONST_srv_state_awaiting_connection) {
			# This is a connection request

			# Build selector for available devices
			my ($device_selector) =
				BuildDeviceSelector( $req_item_id,
					$building_id, $block_number );

			# If there is a block number it gets displayed
			# as part of the room number
			$room_number = "$block_number/$room_number" if ($block_number ne "");

			# Build connection row for the job table
			$job_row = BuildConnectRow( $req_item_id,
					$room_number, $outlet_number,
					$device_selector, $user, $fullname );
		}
		elsif ($service_state_id ==
				$CONST_srv_state_awaiting_disconnection) {
			# This is a disconnection request
			# so we need data from the existing connection

			my ($building_id, $building_name, $block_number,
				$room_number, $outlet_number,
				$device_id, $device_name, $device_port,
				$service_request_item_id,
				$date_connected, $date_disconnected,
				$connected_by, $disconnected_by,
				$service_state_id, $service_state_description,
				$comment) = FetchConnection( "%", $req_item_id);

			# If there is a block number it gets displayed
			# as part of the room number
			$room_number = "$block_number/$room_number" if ($block_number ne "");

			# Build connection row for the job table
			$job_row = BuildDisconnectRow( $req_item_id,
					$room_number, $outlet_number,
					$device_name, $device_port,
					$user, $fullname );
		}
		else {
			# Something odd here!
			EmitError( "Internal error: unexpected service_state_id $service_state_id in service_request_item $req_item_id" );
		Log( "err", "unexpected service_state_id $service_state_id in service_request_item $req_item_id" );
			exit 1;
		}


		# Fill in the job details
		$job_hash_array->[$job_ser++] = {
			"row" => $job_row
		};
	}
	$template->param( "cgi_url", $cgi_url );
	$template->param( "buildings", \@bldg_hash_array );

	EmitPage( "Network Connection Jobs",
		"List of connections and disconnections to be made",
		$template->output()
	);

	Log( "info", "network connection joblist" );
}

########################################################################
# Emit a Connection Checklist
########################################################################
sub EmitConnectionCheckList {

	# Get the building ID
	my ($req_building_id) = $query->param("building");
	Trim($req_building_id);
	if (!($req_building_id =~ /^[0-9]+$/)) {
		EmitError( "You must select a building" );
		exit 1;
	}

	# Allow requester to limit display by block/flat
	my ($block_spec) = $query->param("block_spec");
	Trim($block_spec);
	$block_spec = "%" if (!defined($block_spec) || ($block_spec eq ""));

	my ($this_building) = FetchBuildingName( $req_building_id );

	my ($template) = HTML::Template->new(
		filename => "$htmldir/connection-checklist.html" );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );


	my ($list_items) = FetchConnectionList($req_building_id, $block_spec);
	if (!$list_items) {
		EmitError( "FetchConnectionList failed $ConnDB::errstr" );
		Log( "err", "FetchConnectionList failed: $ConnDB::errstr" );
		return 0;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my ($block_ser) = 0;
	my (@block_hash_array);
	my ($connection_hash_array);
	my ($connection_ser);
	my ($last_block_id) = "NOBLOCK";
	for ($item=0;defined($list_items->[$item]);$item++) {
		my ($connection_id, $building_id, $building_name, $block_number,
			$room_number, $outlet_number, $device_id, $device_name,
			$device_port, $service_request_item_id, $date_connected,
			$connected_by, $username) = @{$list_items->[$item]};

		# If this is a new block then we start a new array
		if ($last_block_id ne $block_number) {
			# Create a connection array for the new block
			$connection_ser = 0;
			# This weird construct gets us a reference to a new array:
			$connection_hash_array = \@{[]};
			$block_hash_array[$block_ser] = {
				"building_name" => $building_name,
				"block" => $block_number,
				"items" => $connection_hash_array
			};
			$block_ser++;
			$last_block_id = $block_number;
		}

		# If there is a block number it gets displayed
		# as part of the room number
		$room_number = "$block_number/$room_number" if ($block_number ne "");


		# Fill in the connection details
		$connection_hash_array->[$connection_ser++] = {
			"outlet_number" => $outlet_number,
			"room_number" => $room_number,
			"device_name" => $device_name,
			"device_port" => $device_port,
			"date_connected" => HumaniseDate(substr( $date_connected, 0, 10 )),
			"connected_by" => $connected_by,
			"username" => $username,
			"cgi_url" => $cgi_url
		};
	}
	$template->param( "cgi_url", $cgi_url );
	$template->param( "building_name", $this_building );
	$template->param( "date", HumaniseDate(FetchDBDate) );
	$template->param( "number_of_connections", $item );
	$template->param( "blocks", \@block_hash_array );

	EmitPage( "Connection Checklist for $this_building",
		"List of connections current in $this_building",
		$template->output()
	);

	Log( "info", "connection checklist for $this_building" );
}

########################################################################
# MailConnectedMessage( user, Building, block_number, room_number, outlet )
#
# Inform user by mail that their connection has been made
########################################################################
sub MailConnectedMessage {
	my ($user, $bldg, $block, $room, $outlet) = @_;

	my ($mail_template) = HTML::Template->new(
		filename => "$htmldir/mail-connection-done.txt" );

	$mail_template->param( "building", $bldg );
	$mail_template->param( "block", $block );
	$mail_template->param( "room", $room );
	$mail_template->param( "outlet", $outlet );
	$mail_template->param( "user", $user );

	my ($recipient) = $user;
	if ($TestMode) {
		$recipient = $TestRecipient;
	}

	# Inform user by mail
	SendMail( $recipient, 'Halls Network Connection', $mail_template->output() );
}


########################################################################
# Update the connections database and issue confirmation page
########################################################################
sub UpdateConnections {
	my (@params) = $query->param;

	if (!Permitted( $loggedinuser, "connect" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for connect" );
		exit 1;
	}

	# We want to know who did the work!
	my ($whodidit) = $query->param("connection-operator");
	Trim($whodidit);
	if (!defined($whodidit) || ($whodidit eq "")) {
		EmitError( "You have not indicated who did the work" );
		exit 1;
	}

	if (!FetchFullName($whodidit, 0)) {
		EmitError( "Please use a valid username to indicate who did the work: '$whodidit' is not valid" );
		exit 1;
	}

	my (@update_list);
	# my ($ins) = "<BR>";

	# Work through the params supplied looking for valid device-specs
	# or disconnect-requests
	my ($count) = 0;
	foreach (@params) {
		next if !/^(devsel|disconnect)-([0-9]+)$/;
		# Grab the operation and ID from the two parts of the regex
		my($keyword) = $1;
		my($req_id) = $2;
		$keyword = "connect" if ($keyword eq "devsel");

		$update_list[$count] = [ $keyword, $req_id,
			$query->param("devsel-$req_id"),
			$query->param("port-$req_id"),
			$query->param("outlet-$req_id") ];

		Trim( $update_list[$count][1] );
		Trim( $update_list[$count][2] );
		Trim( $update_list[$count][3] );
		Trim( $update_list[$count][4] );

		# $ins .= join ":", @{$update_list[$count]};
		# $ins .= "<BR>";
		$count++;
	}

	# EmitError( "Found these: $ins" );

	# We now have an array containing one sub-array for each line-item
	# so the next job is to check that they are valid updates
	# While doing this we will store service_request_item data for each
	# valid update in @req_items

	my (@req_items);
	my ($number_to_update) = 0;

	for($count=0; defined($update_list[$count]); $count++) {
		my ($keyword) = $update_list[$count][0];
		my ($req_item_id) = $update_list[$count][1];
		my ($device_id) = $update_list[$count][2];
		my ($port_number) = $update_list[$count][3];
		my ($outlet) = $update_list[$count][4];

		# Ignore unmodified lines
		if ( ($keyword eq "connect") &&
				($device_id eq "") && ($port_number eq "")) {
			# Wipe out the keyword
			$update_list[$count][0] = "";
			next;
		}

		# Get the full details for this service_request_item
		my (@req_item) = FetchRequestItem( $req_item_id );
		$req_items[$count] = \@req_item;

		if (!defined($req_items[$count][3])) {
			EmitError( "Internal error: cannot find service_request_item '$req_item_id'" );
			exit 1;
		}

		# Do we have all the data we require for this item?
		if (($keyword eq "connect") &&
			(($device_id eq "") || ($port_number eq "") || ($outlet eq ""))
			) {
			# No - build an error message and give up
			my ($building) = $req_items[$count][2];
			my ($room) = $req_items[$count][4];
			if ($req_items[$count][3] ne "") {
				$room = "$req_items[$count][3]/$room";
			}
			EmitError( "Missing information for $building $room" );
			exit 1;
		}

		# Has this item already been done?
		if (($keyword eq "connect") &&
				($req_items[$count][9] != $CONST_srv_state_awaiting_connection)) {
			# Yes - build an error message and give up
			my ($building) = $req_items[$count][2];
			my ($room) = $req_items[$count][4];
			if ($req_items[$count][3] ne "") {
				$room = "$req_items[$count][3]/$room";
			}
			EmitError( "The connection for $building $room outlet $outlet has already been processed" );
			exit 1;
		}

		# Same check for deletions
		if (($keyword eq "disconnect") &&
				($req_items[$count][9] != $CONST_srv_state_awaiting_disconnection)) {
			# Yes - build an error message and give up
			my ($building) = $req_items[$count][2];
			my ($room) = $req_items[$count][4];
			if ($req_items[$count][3] ne "") {
				$room = "$req_items[$count][3]/$room";
			}
			EmitError( "The disconnection of $building $room outlet $outlet has already been processed" );
			exit 1;
		}

		# Looks like an OK item
		$number_to_update++;
	}

	# Is there any work left after all that?
	if ($number_to_update < 1) {
		EmitError( "No connection updates were specified" );
		exit 1;
	}

	# So far so good - we have at least one good-looking update
	# and no obvious errors

	# Trawl through the list again and do the updates
	my ($num_connected) = 0;
	my ($num_disconnected) = 0;
	for($count=0; defined($update_list[$count]); $count++) {
		my ($keyword) = $update_list[$count][0];
		my ($req_item_id) = $update_list[$count][1];
		my ($device_id) = $update_list[$count][2];
		my ($port_number) = $update_list[$count][3];
		my ($outlet) = $update_list[$count][4];

		if ($keyword eq "connect") {
			my ($addstate) = AddConnection(
				$req_items[$count][1],	# building_id
				$req_items[$count][3],	# block_number
				$req_items[$count][4],	# room_number
				$outlet,		# Use our own $outlet in case it was changed
				$device_id,
				$port_number,
				$req_item_id,
				$whodidit );

			if (!$addstate) {
				EmitError( "AddConnection failed: $ConnDB::errstr" );
				Log( "err", "AddConnection failed: $ConnDB::errstr" );
				exit 1;
			}

			# Connection processed OK
			$num_connected++;

			# Inform user by mail
			MailConnectedMessage(
				$req_items[$count][12],	# user
				$req_items[$count][2],	# Building
				$req_items[$count][3],	# block_number
				$req_items[$count][4],	# room_number
				$outlet			# Use our own $outlet in case it was changed
			);
		}
		elsif ($keyword eq "disconnect") {
			my ($dis_state) = DeleteConnection($req_item_id, $whodidit);

			if (!$dis_state) {
				EmitError( "DeleteConnection failed: $ConnDB::errstr" );
				Log( "err", "DeleteConnection failed: $ConnDB::errstr" );
				exit 1;
			}

			$num_disconnected++;
		}
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/connection-done.html" );

	$template->param( "num_connected", $num_connected );
	$template->param( "num_disconnected", $num_disconnected );
	$template->param( "cgi_url", $cgi_url );

	EmitPage( "Connections Updated",
		"The connection database has been updated",
		$template->output()
	);
}

########################################################################
#
# CancelItem
#
# Someone has requested that a service_request_item be cancelled
#
########################################################################
sub CancelItem {

	# Get the parameters to the request and check them
	my ($item_id) = $query->param("item_id");
	Trim($item_id);
	if (!defined($item_id) || !($item_id =~ /^[0-9]+$/)) {
		EmitError( "Bad URL or form: no item_id was specified" );
		exit 1;
	}

	my ($confirm) = $query->param("confirm");
	Trim($confirm);
	$confirm =~ tr/A-Z/a-z/;
	if (!defined($confirm) || ($confirm ne "cancel")) {
		EmitError( "If you really want to cancel a network request item you must write \"cancel\" in the confirmation box" );
		exit 1;
	}

	my ($reason) = $query->param("reason");
	Trim($reason);
	if (!defined($reason) || !($reason =~ /[a-zA-Z]+/)) {
		EmitError( "You must give a reason for cancelling the request" );
		exit 1;
	}

	# OK - we have some possible-looking parameters
	# Apply some more sanity checks
	if (!ValidateInputValue("service_request_item", "service_request_item_id",
			$item_id)) {
		EmitError( "No requests found with item ID $item_id" );
		exit 1;
	}

	# Try to do it
	my ($cancelstat) = CancelRequestItem( $item_id, $loggedinuser,
					"Cancelled by $loggedinuser: $reason", 0 );
	if ($cancelstat == 1) {
		EmitError( "You ($loggedinuser) do not have permission to cancel that request item" );
		Log( "err", "permission denied for cancel $item_id" );
		exit 1;
	}
	elsif ($cancelstat == 2) {
		EmitError( "That item cannot be cancelled" );
		exit 1;
	}
	elsif ($cancelstat == 3) {
		EmitError( "Internal error: $ConnDB::errstr" );
		Log( "err", "CancelRequestItem failed: $ConnDB::errstr" );
		exit 1;
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/item-cancelled.html" );

	$template->param( "date", HumaniseDate(FetchDBDate) );

	EmitPage( "Item Cancelled",
		"Network Service request has been cancelled",
		$template->output()
	);

}

########################################################################
# Emit a "network admin" page
########################################################################
sub EmitNetworkAdminPage {

	my ($template) = HTML::Template->new(
		filename => "$htmldir/network-admin.html" );

	$template->param( "cgi_url", $cgi_url );

	# We want all known buildings on the building list
	my ($building_selector) = BuildBuildingSelector( "%" );
	$template->param( "building_selector", $building_selector );

	my ($site_selector) = BuildSiteSelector;
	$template->param( "site_selector", $site_selector );

	EmitPage( "Network Admin",
		"Network connection administration: add and modify devices etc",
		$template->output() );
	Log( "info", "network admin page" );
}

########################################################################
# Add a building
########################################################################
sub AddBuilding {

	if (!Permitted( $loggedinuser, "connect" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for add building" );
		exit 1;
	}

	# Get the site_id
	my ($site_id) = $query->param("site_id");
	Trim($site_id);
	if (!defined($site_id) || ($site_id < 0)) {
		EmitError( "You must specify a site" );
		exit 1;
	}
	if (!ValidateInputValue("sites", "site_id", $site_id)) {
		EmitError( "Unknown site_id specified" );
		exit 1;
	}

	# Building name
	my ($building_name) = $query->param("name");
	Trim($building_name);
	if (!defined($building_name) || ($building_name eq "")) {
		EmitError( "You must give the building name" );
		exit 1;
	}
	if (!OKAlphaNumPunct($building_name)) {
		EmitError( "Bad characters in building name" );
		exit 1;
	}
	if (ValidateInputValue( "buildings", "building_name", $building_name )) {
		EmitError( "$building_name is already listed in the database" );
		exit 1;
	}

	# Building code
	my ($building_code) = $query->param("code");
	Trim($building_code);
	if (!defined($building_code) || ($building_code eq "")) {
		EmitError( "You must give the building code" );
		exit 1;
	}
	if (!OKAlphaNumPunct($building_code)) {
		EmitError( "Bad characters in building code" );
		exit 1;
	}
	# Uppercase the building codes
	$building_code =~ tr/a-z/A-Z/;
	if (ValidateInputValue( "buildings", "building_code", $building_code )) {
		EmitError( "Code $building_code is already listed in the database" );
		exit 1;
	}

	# Building type
	my ($building_type) = $query->param("building-type");
	Trim($building_type);
	if (!defined($building_type) || ($building_type eq "")) {
		EmitError( "You must specify the building type" );
		exit 1;
	}
	if (!ValidateInputValue("building_types", "building_type", $building_type)) {
		EmitError( "Unknown building type" );
		exit 1;
	}

	# Does this building have flats/blocks?
	my ($has_blocks) = $query->param("has-blocks");
	Trim($has_blocks);
	if (defined($has_blocks) && ($has_blocks ne "")) {
		$has_blocks = 1;
	}
	else {
		$has_blocks = 0;
	}

	# Request looks OK - try to process it
	if (!AddBuildingToDB( $building_code, $building_name, $building_type,
				$has_blocks, $site_id)) {
		EmitError( "AddBuildingToDB failed: $ConnDB::errstr" );
		Log( "err", "AddBuildingToDB failed: $ConnDB::errstr" );
		exit 1;
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/added.html" );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	my ($text) = "Added $building_name to the database";
	$template->param( "text", $text );
	$template->param( "cgi_url", $cgi_url );
	EmitPage( "Added", $text, $template->output() );
}

########################################################################
# Add a device
########################################################################
sub AddDevice {

	if (!Permitted( $loggedinuser, "connect" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for add device" );
		exit 1;
	}

	# Get the building_id
	my ($building_id) = $query->param("building");
	Trim($building_id);
	if (!defined($building_id) || ($building_id < 0)) {
		EmitError( "You must specify a building" );
		exit 1;
	}
	if (!ValidateInputValue("buildings", "building_id", $building_id)) {
		EmitError( "Unknown building_id specified" );
		exit 1;
	}

	# Building name
	my ($device_name) = $query->param("device_name");
	Trim($device_name);
	if (!defined($device_name) || ($device_name eq "")) {
		EmitError( "You must give the device name" );
		exit 1;
	}
	if (!OKAlphaNumPunct($device_name)) {
		EmitError( "Bad characters in device name" );
		exit 1;
	}
	if (ValidateInputValue( "devices", "device_name", $device_name )) {
		EmitError( "$device_name is already listed in the database" );
		exit 1;
	}

	# Flat/block number
	my ($block) = $query->param("block");
	Trim($block);
	if (defined($block) && !OKAlphaNum($block)) {
		EmitError( "Bad characters in flat/block number" );
		exit 1;
	}
	$block = "" if !defined($block);
	FixCase($block);
	if (($block eq "") && HasBlocks($building_id)) {
		EmitError( "That building has flats or blocks but you have not filled in the flat/block number" );
		exit 1;
	}
	if (($block ne "") && !HasBlocks($building_id)) {
		EmitError( "You have filled in a flat/block number but the building does not have flats or blocks." );
		exit 1;
	}


	# Get the device_type
	my ($device_type) = $query->param("device_type");
	Trim($device_type);
	if (!defined($device_type) || !($device_type =~ /^[0-9]+$/)) {
		EmitError( "You must specify a device_type" );
		exit 1;
	}

	# Request looks OK - try to process it
	if (!AddDeviceToDB( $device_name, $building_id,
				$block, $device_type)) {
		EmitError( "AddDeviceToDB failed: $ConnDB::errstr" );
		Log( "err", "AddDeviceToDB failed: $ConnDB::errstr" );
		exit 1;
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/added.html" );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	my ($text) = "Added $device_name to the database";
	$template->param( "text", $text );
	$template->param( "cgi_url", $cgi_url );
	EmitPage( "Added", $text, $template->output() );
}


########################################################################
# Load connection data (supplied as CSV file upload)
########################################################################
sub LoadOutlets {

	if (!Permitted( $loggedinuser, "connect" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for load connection data" );
		exit 1;
	}

	# Get the building_id
	my ($building_id) = $query->param("building");
	Trim($building_id);
	if (!defined($building_id) || ($building_id < 0)) {
		EmitError( "You must specify a building" );
		exit 1;
	}
	if (!ValidateInputValue("buildings", "building_id", $building_id)) {
		EmitError( "Unknown building_id specified" );
		exit 1;
	}

	# Get the filename (may not need this)
	my ($uploaded_filename) = $query->param("outlet-data-file");
	if (!defined($uploaded_filename)) {
		EmitError( "You must specify a file to be loaded" );
		exit 1;
	}


	# Get handle on uploaded file
	my ($upload_FH) = $query->upload("outlet-data-file");
	if (!defined($upload_FH)) {
		EmitError( "Bad filename" );
		exit 1;
	}

	my ($delete_flag) = $query->param("delete");
	Trim($delete_flag);
	$delete_flag = (defined($delete_flag) && ($delete_flag eq "delete"));

	# Request looks OK - try to process it

	# Find out whether this building has blocks/flats
	my ($BldgHasBlocks) = HasBlocks($building_id);

	## my ($message) = "<PRE>\n";

	# Read the whole file, check every line, and store in array
	my ($item) = 0;
	my (@outlets);
	while(<$upload_FH>) {
		# Get rid of line-end chars (including DOS-ish ones)
		s/[\r\n]*//g;
		# Ignore comments
		next if /^\s*#/;
		# Ignore blank lines
		next if /^\s*$/;

		# Magic from Text::ParseWords for CSV parsing:
		my ($block, $room, $outlet, $comment) = parse_line( ",", 0, $_);
		# Comments are not essential, but must not be null
		$comment = "" if !defined($comment);

		Trim($block); Trim($room); Trim($outlet); Trim($comment);
		FixCase($block);

		## $message .= "block=#$block# room=#$room# outlet=#$outlet# comment=#$comment#\n";


		# Each line must have at least three fields
		if (!defined($outlet)) {
			EmitError( "Format error in line: '$_'" );
			exit 1;
		}

		# A block number must be given IFF building has blocks
		if (($block eq "") && $BldgHasBlocks) {
			EmitError( "Missing block/flat number in line: '$_'" );
			exit 1;
		}
		if (($block ne "") && !$BldgHasBlocks) {
			EmitError( "Unwanted block/flat number in line: '$_'" );
			exit 1;
		}

		# Looks good - store it!
		$outlets[$item++] = [ $block, $room, $outlet, $comment ];
	}
	## $message .= "</PRE>";

	# The file checks out OK, so load it into the database
	my ($res) = LoadOutletData( $building_id, \@outlets, 1, $delete_flag );
	my ($NLoaded, $NDuplicates, $NDeleted) = ($res->[0], $res->[1], $res->[2]);
	if ($NLoaded < 0) {
		EmitError( "Internal error: LoadOutletData failed: $ConnDB::errstr" );
		exit 1;
	}

	# All OK so tell the user
	my ($template) = HTML::Template->new(
		filename => "$htmldir/loaded-outlets.html" );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	$template->param( "nloaded", $NLoaded );
	$template->param( "nduplicates", $NDuplicates );
	$template->param( "ndeleted", $NDeleted );

	$template->param( "cgi_url", $cgi_url );
	EmitPage( "Loaded outlet data", "Loaded outlet data", $template->output() );
}

########################################################################
# Emit a page showing current connection stats
#
########################################################################
sub ShowStats {

	# We should really get the state names from the database
	# but this is a quick hack...
	my ($statsline) = GetCounts;
	my ($N0, $N1, $N2, $N3, $N4, $N5, $N6, $N7) = split /\s+/, $statsline;

	my ($template) = HTML::Template->new(
		filename => "$htmldir/stats-page.html" );

	$template->param( "N0", $N0 );
	$template->param( "N1", $N1 );
	$template->param( "N2", $N2 );
	$template->param( "N3", $N3 );
	$template->param( "N4", $N4 );
	$template->param( "N5", $N5 );
	$template->param( "N6", $N6 );
	$template->param( "N7", $N7 );

	$template->param( "totconn", $N3+$N4+$N7 );

	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	EmitPage( "Network connection stats",
		"Network connection stats",
		$template->output() );
	Log( "info", "list connection stats" );
}

########################################################################
# Emit a list of outlets for a particular location
#
########################################################################
sub ListOutlets {

	# We must have a valid building
	my($building_id) = $query->param("building");
	Trim($building_id);
	if (!ValidateInputValue("buildings", "building_id", $building_id)) {
		EmitError( "You must select a building" );
		exit 1;
	}

	my ($building_name) = FetchBuildingName( $building_id );
	if (!$building_name) {
		# This should not be possible as we verified the building above!
		EmitError( "Internal error: missing building description" );
		exit 1;
	}

	# There may be a block number
	my($block) = $query->param("block");
	Trim($block);
	if (defined($block) && !OKAlphaNum($block)) {
		EmitError( "Bad characters in flat/block number" );
		exit 1;
	}
	$block = "" if !defined($block);

	# There may be a room number
	my($room) = $query->param("room");
	Trim($room);
	if (defined($room) && !OKAlphaNumPunct($room)) {
		EmitError( "Bad characters in room number" );
		exit 1;
	}
	$room = "%" if (!defined($room) || ($room eq ""));

	if (($block eq "") && HasBlocks($building_id)) {
		EmitError( "$building_name has flats or blocks but you have not filled in the flat/block number" );
		exit 1;
	}

	if (($block ne "") && !HasBlocks($building_id)) {
		EmitError( "You have filled in a flat/block number but $building_name does not have flats or blocks." );
		exit 1;
	}

	my ($outlets) = FetchOutlets( $building_id, $block, $room );

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my (@item_hash_array);
	for ($item=0;defined($outlets->[$item]);$item++) {
		$item_hash_array[$item] = {
			"room_number" => $outlets->[$item][0],
			"outlet_number" => $outlets->[$item][1]
		}
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/list-outlets.html" );

	my ($blockspec) = "";
	$blockspec = "block/flat $block" if ($block ne "");
	$blockspec .= " room $room" if ($room ne "%");
	$template->param( "blockspec", $blockspec );

	$template->param( "building_name", $building_name );
	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );
	$template->param( "number_of_outlets", $item );

	$template->param( "outlets", \@item_hash_array );
	$template->param( "cgi_url", $cgi_url );

	EmitPage( "Network Outlets in $building_name $blockspec",
		"List of network outlets in $building_name $blockspec",
		$template->output() );
	Log( "info", "list network outlets" );
}

########################################################################
#
# Emit a list of users with special privileges
#
########################################################################
sub ListPermissions {

	if (!Permitted( $loggedinuser, "admin" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for list permissions" );
		exit 1;
	}

	my ($perms) = FetchPermissionsList();
	if (!$perms) {
		EmitError( "Internal error: unable to fetch permissions list: $ConnDB::errstr" );
		Log( "err", "FetchPermissionsList failed: $ConnDB::errstr" );
		exit 1;
	}

	# Unpick the array of arrays and build an array of hashes
	my ($item);
	my ($usernum) = -1;
	my ($current_user) = "NOSUCHUSER";
	my (@user_hash_array);
	for ($item=0;defined($perms->[$item]);$item++) {
		if ($perms->[$item][0] ne $current_user) {
			$usernum++;
			$current_user = $perms->[$item][0];
			$user_hash_array[$usernum]{"user"} = $current_user;
			# These are the possible privs, and they all start empty:
			$user_hash_array[$usernum]{"admin"} = "";
			$user_hash_array[$usernum]{"billing"} = "";
			$user_hash_array[$usernum]{"connect"} = "";
			$user_hash_array[$usernum]{"fullname"} = $perms->[$item][2];
		}

		# Now select the appropriate privs for this user
		my ($priv) = $perms->[$item][1];
		$user_hash_array[$usernum]{ $priv } = "yes";
	}

	my ($template) = HTML::Template->new(
		filename => "$htmldir/list-permissions.html" );

	$template->param( "permissions", \@user_hash_array );
	my ($date) = FetchDBDate;
	$template->param( "date", HumaniseDate($date) );

	$template->param( "cgi_url", $cgi_url );

	EmitPage( "Permissions",
		"List of who may do what in the connections system",
		$template->output() );
	Log( "info", "permissions page" );
}


########################################################################
#
# UpdatePermissions
#
########################################################################
sub UpdatePermissions {

	if (!Permitted( $loggedinuser, "admin" )) {
		EmitError( "Permission denied" );
		Log( "err", "permission denied for update permissions" );
		exit 1;
	}

	my ($user) = $query->param("user");
	Trim($user);
	if (!defined($user) || ($user eq "")) {
		EmitError( "You must give a user name to add or change" );
		exit 1;
	}
	if (!OKAlphaNumPunct($user)) {
		EmitError( "Bad characters in user name" );
		exit 1;
	}

	# Lowercase the username
	$user =~ tr/A-Z/a-z/;

	if (!FetchFullName( $user, 0 )) {
		EmitError( "Unknown user '$user'" );
		exit 1;
	}

	my ($has_admin) = $query->param("admin");
	$has_admin = 0 if (!defined($has_admin) || ($has_admin eq ""));

	my ($has_billing) = $query->param("billing");
	$has_billing = 0 if (!defined($has_billing) || ($has_billing eq ""));

	my ($has_connect) = $query->param("connect");
	$has_connect = 0 if (!defined($has_connect) || ($has_connect eq ""));

	my ($delete_user) = $query->param("delete_user");
	$delete_user = 0 if (!defined($delete_user) || ($delete_user eq ""));

	# Sanity check
	if (($user eq $loggedinuser) && !$has_admin) {
		EmitError( "You cannot remove your own Admin ability!" );
		exit 1;
	}

	if (!$has_admin && !$has_billing && !$has_connect && !$delete_user) {
		EmitError( "You must check the `delete user' box if you want to remove all permissions" );
		exit 1;
	}

	if ($delete_user && ($has_admin || $has_billing || $has_connect)) {
		EmitError( "You cannot add permissions at the same time as deleting the same user" );
		exit 1;
	}

	# Add the user to the database if not already there
	if (!FetchFullName($user, 1)) {
		if (!AddUser($user)) {
			EmitError( "Internal error: cannot add user $user to the database." );
			Log( "err", "AddUser($user) failed" );
			exit 1;
		}
	}

	SetPermissions( $user, $has_admin, $has_billing, $has_connect );

	# Redisplay to show new settings
	ListPermissions;
}

########################################################################
# Process a CGI request
########################################################################

sub HandleCGI {
	# Prepare to get data from CGI submission
	$query = new CGI;

	if ($SaveQueries) {
		open QUERYFILE, ">>$SaveQueries" || die "cannot open queryfile";
		$query->save(\*QUERYFILE);
		close QUERYFILE;
	}


	$mode = $query->param("mode");
	Trim($mode);
	if (!defined($mode) || ($mode ne "halls")) {
		EmitError( "Bad URL or form: no valid mode was specified" );
		exit 1;
	}

	$op = $query->param("op");
	Trim($op);
	if (!defined($op)) {
		EmitError( "Bad URL or form: no operation was specified" );
		exit 1;
	}

	$loggedinuser = $query->remote_user;
	if ($AllowFudgeUser && !defined($loggedinuser)) {
		# This is for testing without a Web server
		$loggedinuser = $query->param("webuser");
	}
	Trim($loggedinuser);

	if (!defined($loggedinuser)) {
		EmitError( "Must be authenticated to use this service" );
		Log( "err", "unauthenticated user" );
		exit 1;
	}

	# We now know the name of the user so tell the log system about it
	my($remote_host) = $query->remote_host();
	$remote_host = "UNKNOWN" if !defined($remote_host);
	my($auth_type) = $query->auth_type();
	$auth_type = "UNKNOWN" if !defined($auth_type);
	LogUser( "$loggedinuser\@$remote_host/$auth_type" );

	# What name have we been called under?
	my ($script_name) = $query->script_name();
	Log( "info", "Script name $script_name" );


	# Wind up the database
	if (Connect( $database_param_file )) {
		EmitError "Connect failed $ConnDB::errstr";
		Log( "err", "Connect failed $ConnDB::errstr" );
		exit 1;
	}

	if    ($op eq "payment-list") { EmitPaymentList }
	elsif ($op eq "accept-payment") { AcceptPayment }
	elsif ($op eq "connection-list") { EmitConnectionList }
	elsif ($op eq "connection-checklist") { EmitConnectionCheckList }
	elsif ($op eq "user-account") { EmitUserAccount }
	elsif ($op eq "proxy-apply-1") { EmitProxyApplicationPage }
	elsif ($op eq "apply-halls-1") { EmitApplyForHallsServicePage }
	elsif ($op eq "apply-halls-2") { EmitApplyForHallsPaymentPage(2) }
	elsif ($op eq "proxy-apply-2") { EmitApplyForHallsPaymentPage(2) }
	elsif ($op eq "apply-halls-3") { EmitApplyForHallsPaymentPage(3) }
	elsif ($op eq "renew-item") { EmitHallsRenewalPaymentPage(2) }
	elsif ($op eq "renew-halls-3") { EmitHallsRenewalPaymentPage(3) }
	elsif ($op eq "update-connections") { UpdateConnections }
	elsif ($op eq "cancel-item") { CancelItem }
	elsif ($op eq "network-admin") { EmitNetworkAdminPage }
	elsif ($op eq "add-building") { AddBuilding }
	elsif ($op eq "add-device") { AddDevice }
	elsif ($op eq "list-outlets") { ListOutlets }
	elsif ($op eq "list-permissions") { ListPermissions }
	elsif ($op eq "update-permissions") { UpdatePermissions }
	elsif ($op eq "load-outlet-data") { LoadOutlets }
	elsif ($op eq "stats") { ShowStats }
	elsif ($op eq "show") { ShowParams }
	elsif ($op eq "error") { EmitError( "mode=$mode, op=$op, loggedinuser=$loggedinuser" ) }
	else {
		EmitError( "Bad URL or form: unknown operation '$op' was specified" );
		Log( "err", "unknown operation '$op' was specified" );
		exit 1;
	}

	return 0;
}

########################################################################
# End of ConnWeb.pm
########################################################################

1;

########################################################################
########################################################################
########################################################################

